#ifndef	lint
static char sccsid[] = "@(#)avl.c	1.12 29 Sep 1993";
#endif

/* This software is free: please copy it and/or modify it. Do with it
what you like.  It is a gift from KnowledgeWare Inc, and was written by
Tom Howland.

An AVL tree is an ordered tree with the following characteristics.

- each node (subtree) points to two node (it is a binary tree).

- for every node, the heights of its two subtrees differ by at most 1
  (this is the balance constraint).

During node insertion and deletion, the balance constraint may
temporarily be violated.  A tree is rebalanced by a process called
rotation.  There are two categories of rotation: single and double.

Here is an example of a single rotation.

  Before Insertion   After Insertion     After Rotation
         |                |                      |
         5                5                      3
       /   \            /   \                  /   \
       3   6            3   6                 2    5
      / \              / \                   /    / \
      2 4              2 4                   1    4 6
                      /
                      1

The tree was balanced before the insertion.  After the insertion, the
node with value 5 is not balanced (its left subtree has a depth of 3,
while its right subtree has a depth of 1).  The balance is restored by a
single rotation at that node.

Here is an example of a double rotation.

  Before Insertion   After Insertion     After Rotation
         |                |                      |
         5                5                      4
       /   \            /   \                  /   \
       2   6            2   6                 2    5
      / \              / \                   / \    \
      1 4              1 4                   1 3    6
                        /
                        3

The tree was balanced before the insertion.  After the insertion, the
node with value 5 is not balanced (its left subtree has a depth of 3,
while its right subtree has a depth of 1).  The balance is restored by a
double rotation at that node.

I, Tom Howland, hacked this very much to work with Prolog.  I worked
from some files I got from Vick Bancroft of KnowledgeWare, who got it
presumably from Tony Bolmarcich (TONYB at YKTVMH) and Tim Bell (BELLTG
at WINVMB) who I suspect work for IBM and translated it from a PL/I
source.  Presumably they got it from Niklaus Wirth, "Algorithms and Data
Structures", Prentice-Hall, 1986, but AVL trees were originally
described in: G.M.  Adelson-Velskii and E.M.  Landis, Doklady Akademia
Nauk, SSSR, 146(1962), 263-266; English translation in Soviet Math, 3,
1259-1263.  In any case this implementation bears little resemblence to
the original and the mistakes here are mine.

This code could be optimized. I had to add "parent pointers" to it so we
could do walks from Prolog. Having parent pointers means that we don't
have to rely on recursion to keep our place during rebalancing.  So all
of these functions could be rewritten so as to not rely on recursion.
Function calls are expensive in C (Prolog procedure calls are much
faster -- seriously). I tried rewriting "delete()" in an interative
manner. I could never get it quite right: you're welcome to try.

Tom Howland, April 1993 */

#include <memory.h>
#include <quintus/quintus.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef int boolean;
#define   true   ((boolean) -1)
#define   false  ((boolean) 0)

typedef int side_t;

/* index value for node pointer array */
#define LEFT ((side_t) 0)
#define RIGHT ((side_t) 1)
/* return opposite side */
#define opp_side(s) ((side_t)(!s))
 
/* additional value for the deeper component of the node link */
#define EQUAL_DEPTH ((side_t) 2)

typedef struct node{
  struct node *parent;	 /* parent of this node. */
  struct node *p[2];	 /* left and right child nodes */
#if 0
  struct node *del_next; /* next node deleted during walk */
#endif
  side_t       deeper;	 /* side with deeper subtree */
  QP_db_reference it;
} nt, *np;

typedef struct tree{
  np        root;			/* root node of tree */
  boolean allow_duplicates;
  int count;
#if 0
  boolean terminate;			/* kill the tree after the walk */
  int walk; */
  np postwalk_root;			/* nodes modified during walk */
  np del_first;				/* nodes deleted during walk */
  QP_db_reference tree_3_ref;
#endif
} tt, *tp;

/* structure for insert() to haul around */

typedef struct{
  QP_term_ref order, ref, key, term;
  boolean added_level;
  tp tree;
/*  np postwalk; */
} insert_data, *insert_data_p;

/* structure for find() to haul around */

typedef struct{
  QP_term_ref order, ref, key, term;
} find_data;

/* structure for delete_recurse() to haul around */

typedef struct{
  QP_term_ref order, ref, key, term;
  boolean removed_level;
} delete_data, *delete_data_p;

static QP_pred_ref
  bt_assert,
  bt_instance,
  bt_unify,
  compare_key,
  compare_ref_term,
  compare_ref_term_vg,
  compare_ref_ref,
  erase,
/*  order, */
  ref_key;

static QP_pred_ref get_ref(Pred, Arity, Mod)
  char *Pred;
  int Arity;
  char *Mod;
{
  QP_pred_ref ref = QP_predicate(Pred, Arity, Mod);
  if(ref == QP_BAD_PREDREF){
    QP_fprintf(QP_stderr,
	       "avl_init: could not find %s:%s/%d", Pred, Arity, Mod);
    QP_action(QP_ABORT);
  }
  return ref;
}

/* Prolog calls "avl_init" to set up the above static pointers */

void avl_init()
{
  char *Mod = "tavl";

  bt_assert           = get_ref("bt_assert",           3, Mod);
  bt_instance         = get_ref("bt_instance",         3, Mod);
  bt_unify            = get_ref("bt_unify",            3, Mod);
  compare_key         = get_ref("compare_key",         4, Mod);
  compare_ref_term    = get_ref("compare_ref_term",    5, Mod);
  compare_ref_term_vg = get_ref("compare_ref_term_vg", 5, Mod);
  compare_ref_ref     = get_ref("compare_ref_ref",     4, Mod);
  erase               = get_ref("erase",               1, Mod);
/*  order               = get_ref("order",               2, Mod); */
  ref_key             = get_ref("ref_key",             2, Mod);
}

/* Rebalance the tree after a node insertion adds a level to a subtree. */

static void ins_rebalance(node, side, added_level)
  np        *node;			/* subtree root            */
  side_t     side;			/* which level was added   */
  boolean   *added_level;		/* level was added to tree */
{
  np cur_link;
 
  cur_link = *node;
  if(cur_link->deeper == opp_side(side)){
    cur_link->deeper = EQUAL_DEPTH;
    (*added_level) = false;
  }else if(cur_link->deeper == EQUAL_DEPTH){
    cur_link->deeper = side;
    (*added_level) = true;
  }else{		/* deeper on side added to */
    np p1;
    p1 = cur_link->p[side];
    if(p1->deeper == side){	/* single rotation         */
    
      cur_link->p[side] = p1->p[opp_side(side)];
      p1->p[opp_side(side)] = cur_link;
      
      p1->parent = cur_link->parent;
      cur_link->parent = p1;
      {
        np x;
	x = cur_link->p[side];
	if(x) x->parent = cur_link;
      }
      
      cur_link->deeper = EQUAL_DEPTH;
      (*node) = p1;
    }else{			/* double rotation         */
      np p2;
      p2 = p1->p[opp_side(side)];
      
      p1->p[opp_side(side)] = p2->p[side];
      p2->p[side] = p1;
      cur_link->p[side] = p2->p[opp_side(side)];
      p2->p[opp_side(side)] = *node;
      
      p2->parent = cur_link->parent;
      p1->parent = p2;
      cur_link->parent = p2;
      {
        np x;
        x = cur_link->p[side];
	if(x) x->parent = cur_link;
	x = p1->p[opp_side(side)];
	if(x) x->parent = p1;
      }

      if(p2->deeper == side)
        cur_link->deeper = opp_side(side);
      else
        cur_link->deeper = EQUAL_DEPTH;
      if(p2->deeper == opp_side(side))
        p1->deeper = side;
      else
        p1->deeper = EQUAL_DEPTH;
      (*node) = p2;
    }
    (*node)->deeper = EQUAL_DEPTH;
    (*added_level) = false;
  }
  return;
}

static compare_dt(order, d, r, k, t)
  QP_db_reference *d;
  QP_term_ref order, r, k, t;
{
  int result;
  QP_put_db_reference(r, *d);
  (void) QP_query(compare_ref_term, order, r, k, t, &result);
  return result;
}

static boolean insert();

static boolean maybe_rebalance(n, side, id)
  np *n;
  side_t side;
  insert_data_p id;
{
  boolean inserted = insert(&((*n)->p[side]), *n, id);
  if(id->added_level) ins_rebalance(n, side, &id->added_level);
  return inserted;
}

static boolean insert(node, parent, id)
  np *node;			  /* subtree root            */
  np  parent;
  insert_data_p id;
{
  boolean inserted;

  if(*node){

    int cmp_result = compare_dt(id->order, &(*node)->it, id->ref,
                                id->key, id->term);
 
    if(cmp_result == 1)
    
    	/* add node to left        */

      inserted = maybe_rebalance(node, LEFT, id);

    else if(cmp_result || id->tree->allow_duplicates)

      /* add node to right    */

      inserted = maybe_rebalance(node, RIGHT, id);

    else{

	/* node already in tree    */

      (id->added_level) = false;
      inserted = false;
    }

  }else if(id->tree->allow_duplicates || parent == (np) 0 ||

      /* make sure the parent key does not unify with this one. */

       QP_query(ref_key, id->ref, id->key) != QP_SUCCESS){
       	
	    /* insert as a leaf here   */
    np n;
      
/*    if(n = id->postwalk); */
    if(0);
    else{
      n = (np) QP_malloc(sizeof(struct node));
      (void) QP_query(bt_assert, id->key, id->term, id->ref);
      (void) QP_get_db_reference(id->ref, &n->it);
/*    n->del_next = (np) 0; */
    }

    n->parent = parent;

    n->p[LEFT] = n->p[RIGHT] = (np) 0;
    n->deeper = EQUAL_DEPTH;

    id->added_level = true;
    (*node) = n;
    inserted = true;
    id->tree->count++;

  }else{

    (id->added_level) = false;
    inserted = false;
  }
  return inserted;
}

tp avl_new_tree(allow_duplicates)
  boolean allow_duplicates;
{
  tp x = (tp) QP_malloc(sizeof(struct tree));
  x->allow_duplicates = allow_duplicates;
  x->count = 0;
  x->root = (np) 0;
/*
  x->walk = 0;
  x->terminate = false;
  x->postwalk_root = x->del_first = (np) 0;
*/
  return x;
}

#ifdef debug

#define free_tree(n) memset(n, 0, sizeof(struct tree)), (void) QP_free((char *) n)
#define free_node(n) memset(n, 0, sizeof(struct node)), (void) QP_free((char *) n)

#else

#define free_tree(n) (void) QP_free((char *) n)
#define free_node(n) (void) QP_free((char *) n)

#endif

static void delete_tree(n, ref)
  np n;
  QP_term_ref ref;
{
  while(n){
    np r;
    delete_tree(n->p[LEFT], ref);
    r = n->p[RIGHT];
    QP_put_db_reference(ref, n->it);
    (void) QP_query(erase, ref);
    free_node(n);
    n = r; 
  }
}

void avl_delete_tree(tree)
  tp tree;
{
/*  if(tree->walk) tree->terminate = true; */
  if(0);
  else{
    QP_term_ref ref = QP_new_term_ref();
/*    QP_put_db_reference(ref, tree->tree_3_ref);
    (void) QP_query(erase, ref); */
    delete_tree(tree->root, ref);
    free_tree(tree);
  }
}

/* Each operation requires a different match:

insert | exact on key when no duplicates, otherwise exact on both key and term
delete | unification
find   | a range

*/

static compare_dk(order, d, ref, key)
  QP_db_reference *d;
  QP_term_ref order, ref, key;
{
  int result;
  QP_put_db_reference(ref, *d);
  (void) QP_query(compare_key, order, ref, key, &result);
  return result;
}

/*
static np find_key(n, Order, key, ref)
  np n;
  QP_term_ref key, ref, Order;
{
  while(n) switch(compare_dk(Order, &n->it, ref, key)){
  case -1:
    n = n->p[RIGHT]; break;
  case 0:
    return n;
  case 1:
    n = n->p[LEFT]; break;
  }
  return (np) 0;
}
*/

/* insert a Key and Term in a tree according to the given order. */

avl_insert(tree, Order, Key, Term)
  tp tree;
  QP_term_ref Order, Key, Term;
{
  insert_data id;
  id.ref = QP_new_term_ref();
  id.key = Key;
  id.term = Term;
  id.order = Order;
  id.tree = tree;
/*
  id.postwalk = (np) 0;
  if(tree->walk) {
    if(tree->allow_duplicates);
    else{
      np n = find_key(tree->root, Order, Key, id.ref);
      if(n && n->del_next == (np) 0) return -1;
    }
    return !insert(&tree->postwalk_root, (np) 0, &id);
  }else 
*/
  return !insert(&tree->root, (np) 0, &id);
}

/* instantiate the key, k, and the term, t, given a database reference, i. */

static boolean instance(r, i, k, t)
  QP_term_ref r, k, t;
  QP_db_reference *i;
{
  QP_put_db_reference(r, *i);
  return (QP_query(bt_instance, r, k, t) == QP_SUCCESS);
}

static boolean unify(r, i, k, t)
  QP_term_ref r, k, t;
  QP_db_reference *i;
{
  QP_put_db_reference(r, *i);
  return (QP_query(bt_unify, r, k, t) == QP_SUCCESS);
}

/* given the current position in the tree, find the next position. We do
this by either going to the leftmost node of the right subtree or looking for
the first ancestor that has a left branch pointing in our direction. */

static np next(n)
  np n;
{
  np x;
  if(x = n->p[RIGHT])for(n = x->p[LEFT]; n; x = n, n = n->p[LEFT]);
  else for(x = n->parent; x && x->p[RIGHT] == n;  n = x, x = x->parent);
  return x;
}

np avl_next(n, ref)
  np n;
  QP_term_ref ref;
{
  np x = next(n);
  if(x) QP_put_db_reference(ref, x->it);
  return x;
}

/* Rebalance after a deletion. */

static void del_rebalance(node, side, removed_level)
  np *node;			/* node to rebalance */
  side_t side;			/* side deleted from */
  boolean *removed_level;	/* whether level was removed */
{
  np link;
 
  link = *node;
  if(link->deeper == side){
    link->deeper = EQUAL_DEPTH;
    *removed_level = true;
  }else if(link->deeper == EQUAL_DEPTH){
    link->deeper = opp_side(side);
    *removed_level = false;
  }else{
    np p1 = link->p[opp_side(side)];
    if(p1->deeper == side){			/* double rotation */
    
      np p2 = p1->p[side];
      p1->p[side] = p2->p[opp_side(side)];
      p2->p[opp_side(side)] = p1;
      link->p[opp_side(side)] = p2->p[side];
      p2->p[side] = *node;
     
      *node = p2;

      p2->parent = link->parent;
      p1->parent = p2;
      link->parent = p2;
      {
        np x;
	x = link->p[opp_side(side)];
	if(x) x->parent = link;
	x = p1->p[side];
	if(x) x->parent = p1;
      }

      if(p2->deeper == opp_side(side))
        link->deeper = side;
      else
        link->deeper = EQUAL_DEPTH;
      if(p2->deeper == side)
        p1->deeper = opp_side(side);
      else
        p1->deeper = EQUAL_DEPTH;
      p2->deeper = EQUAL_DEPTH;
      *removed_level = true;
    }else{					/* single rotation */
      link->p[opp_side(side)] = p1->p[side];
      p1->p[side] = link;
      
      p1->parent = link->parent;
      link->parent = p1;
      {
        np x;
	x = link->p[opp_side(side)];
	if(x) x->parent = link;
      }
      if(p1->deeper == EQUAL_DEPTH){
        link->deeper = opp_side(side);
        p1->deeper = side;
        *removed_level = false;
      }else{
        link->deeper = EQUAL_DEPTH;
        p1->deeper = EQUAL_DEPTH;
      }
      *node = p1;
    }
  }
  return;
}

static void replace_with_lone_child(target, source, side)
  np *target;
  np source;
  side_t side;
{
  np x = source->p[side];
  *target = x;
  if(x) x->parent = source->parent;
}

/* Find and unlink the rightmost node from a subtree. */

static void unlink_rightmost(node, rightmost_node, removed_level)
  np *node;			/* subtree root */
  np *rightmost_node;		/* rightmost node that was found */
  boolean *removed_level;		/* whether a level was removed */
{
  np link = *node;

  if(link->p[RIGHT] == (np) 0){
    *rightmost_node = link;
    replace_with_lone_child(node, link, LEFT);
    *removed_level = true;
  }else{
    unlink_rightmost(&link->p[RIGHT], rightmost_node, removed_level);
    if(*removed_level) del_rebalance(node, RIGHT, removed_level);
  }
}

static boolean delete_recurse();

static boolean maybe_del_rebalance(node, s, dd)
  np *node;
  side_t s;
  delete_data_p dd;
{
  boolean found;
  np n = *node;
  found = delete_recurse(&n->p[s], dd);
  if(dd->removed_level) del_rebalance(node, s, &dd->removed_level);
  return found;
}

#ifdef echo

static void echo_tree(n, s, i)
  np n; int i; char *s;
{
  const static char * dp[3] = {"left", "right", "equal"};
  if(n){
    int h = n->it.count;
    (void) QP_printf("%s child of %d is %d, deeper: %s\n", s, i, h, dp[n->deeper]);
    {
      np pn = n->parent;
      if(pn)
	if(pn->it.count == i);
        else QP_printf("!!! parent mismatch: %d\n", pn->it.count);
      else;
    }
    echo_tree(n->p[LEFT], "left", h);
    echo_tree(n->p[RIGHT], "right", h);
  }
}

#endif

static boolean delete_recurse(node, dd)
  np *node;
  delete_data_p dd;
{
  boolean found;
  np link = *node;
          
  if(link) switch(compare_dt(dd->order, &link->it, dd->ref, dd->key, dd->term)){

  case 1:			/* delete from left subtree */

    found = maybe_del_rebalance(node, LEFT, dd);
    break;

  case -1:			/* delete from right subtree */

    found = maybe_del_rebalance(node, RIGHT, dd);
    break;

  case 0:			/* found node to be deleted */

    found = true;

    if(link->p[RIGHT] == (np) 0){
    
      replace_with_lone_child(node, link, LEFT);
      dd->removed_level = true;

    }else if(link->p[LEFT] == (np) 0){

      replace_with_lone_child(node, link, RIGHT);
      dd->removed_level = true;

    }else{

      /* move the rightmost node of the left subtree here, but keep the
	 value of the link fields of the node being deleted. */
	 
      np x;

      unlink_rightmost(&(link->p[LEFT]), node, &dd->removed_level);
      x = *node;
      link->p[RIGHT]->parent = x;
      if(link->p[LEFT]) link->p[LEFT]->parent = x;
      x->p[LEFT] = link->p[LEFT];
      x->p[RIGHT] = link->p[RIGHT];
      x->deeper = link->deeper;
      x->parent = link->parent;

#ifdef echo
      (void) QP_printf("Before rebalance\n");
      echo_tree(x, "subroot", x->it.count);
#endif
      if(dd->removed_level) del_rebalance(node, LEFT, &dd->removed_level);
#ifdef echo
      (void) QP_printf("after rebalance\n");
      echo_tree(*node, "subroot", (*node)->it.count);
#endif
    }

    (void) QP_query(erase, dd->ref);
    (void) free_node(link);

  }else{

    dd->removed_level = false;
    found = false;
  }

  return found;
}

/* compare_ref_ref_please() is a hack. It means that there is something
   wrong with the algorithm for find_range(). What happens is that when
   find_range() is called with a ground term that doesn't exist in the
   database, the range finding algorithm gives un-understood results,
   that can be characterized as getting the limits for the range
   backwards, or something. In any case, this fixes it, but should be
   looked at someday. */

static boolean compare_ref_ref_please(order, left, right)
  QP_term_ref order;
  QP_db_reference *left, *right;
{
  int result;
  QP_term_ref l = QP_new_term_ref();
  QP_term_ref r = QP_new_term_ref();

  QP_put_db_reference(l, *left);
  QP_put_db_reference(r, *right);

  (void) QP_query(compare_ref_ref, order, l, r, &result);

  return result;
}

/* find a match, or something that might be a match under fuller unification.
   We can do this since variables are always considered less under the
   "standard order" for prolog terms. I made bloody sure that variables are
   less under the "-" pattern, so that tree walking will work.
   
   First find_range() looks for the leftmost node that may match, then the
   rightmost.
*/

static boolean find_range(root, order, key, term, ref, left, right)
  np root, *left, *right;
  QP_term_ref order, key, term, ref;
{
  np n, last;
  QP_term_ref r = QP_new_term_ref();
  int reality = 0;
  
  n = root;
  last = (np) 0;
  while(n){
    switch(compare_dt(order, &n->it, ref, key, term)){
    case -1:
      n = n->p[RIGHT];
      break;
    case 0:
    case 1:
      last = n;
      n = n->p[LEFT];
      break;
    }
  }
  if(last){
    *left = last;
    n = root;
    do{
      int result;
      QP_put_db_reference(r, n->it);
      (void) QP_query(compare_ref_term_vg, order, r, key, term, &result);
      switch(result){
      case -1:
      case 0:
	last = n;
	n = n->p[RIGHT];
	break;
      case 1:
	n = n->p[LEFT];
	break;
      }
    }while(n);

    if(compare_ref_ref_please(order, &((*left)->it), &last->it) == 1)
      return false;

    *right = last;
    QP_put_db_reference(ref, (*left)->it);
    return true;
  }
  return false;
}

/* find the first thing that actually unifies and that hasn't already
been deleted. */

static np find(n, fd)
  np n;		/* the root */
  find_data *fd;
{
  np right;
  if(find_range(n, fd->order, fd->key, fd->term, fd->ref, &n, &right)){
    while(/* n->del_next || */ !unify(fd->ref, &n->it, fd->key, fd->term)){
      if(n == right) return (np) 0;
      n = next(n);
    }
    return n;
  }
  return (np) 0;
}

static void delete(n, fd)
  np *n;
  find_data *fd;
{
  delete_data dd;

  dd.ref = fd->ref;
  dd.term = fd->term;
  dd.key = fd->key;
  dd.order = fd->order;
  (void) delete_recurse(n, &dd);
}

avl_delete(tree, Order, Key, Term)
  tp tree;
  QP_term_ref Order, Key, Term;
{
  int ret;

  find_data fd;
  fd.ref = QP_new_term_ref();
  fd.term = Term;
  fd.key = Key;
  fd.order = Order;

#ifdef echo
  QP_printf("\n");
  echo_tree(tree->root, "root", tree->root->it.count);
  (void) QP_flush(QP_stdout);
#endif

#if 0

  if(tree->walk){
    np n;
    if(find(tree->postwalk_root, &fd)){
      delete(&tree->postwalk_root, &fd);
      ret = 0;
    }else if(n = find(tree->root, &fd)){

      /* Add a node to the delete list; the list is in LIFO order with the
	 end node pointing to itself; because the end node points to
	 itself, a zero pointer means it is not in the list. */

      n->del_next = tree->del_first ? tree->del_first : n;
      tree->del_first = n;

      ret = 0;
    }else ret = -1;
  }else

#endif

/* You might look at the following and say: "But he's looking up twice!
First to instantiate the thing and then to delete it!"  I have to
instantiate it (using find()) so that I can look it up unambiguously in
delete(). */

  if(find(tree->root, &fd)){
    delete(&tree->root, &fd);
    tree->count--;
    ret = 0;
  }else ret = -1;
  return ret;
}

avl_find(tree, Order, Key, Term, Ref, left, right)
  tp tree;
  QP_term_ref Order, Key, Term, Ref;
  np *left, *right;
{
  if(find_range(tree->root, Order, Key, Term, Ref, left, right)) return 0;
  else return -1;
}

avl_nth(tree, nth, ref)
  tp tree;
  int nth;
  QP_term_ref ref;
{
  int i;
  np x, n;

  n = (np) 0;
  for(x = tree->root; x; n = x, x = x->p[LEFT]);
  for(i = 1; n && i != nth; i++, n = next(n));
  if(n){
    QP_put_db_reference(ref, n->it);
    return 0;
  }
  return -1;
}

avl_count(tree)
  tp tree;
{
  return tree->count;
}

#if 0

This stuff doesn't work under solaris 2.2

static void insert_subtree(target, n, id)
  np *target;
  np n;
  insert_data_p id;
{
  while(n){
    np right;
    insert_subtree(target, n->p[LEFT], id);
    right = n->p[RIGHT];
    id->postwalk = n;
    (void) instance(id->ref, &n->it, id->key, id->term);
    (void) insert(target, (np) 0, id);
    n = right;
  }
}

void avl_save_tree_ref(tree, ref)
  tp tree;
  QP_term_ref ref;
{
  (void) QP_get_db_reference(ref, &tree->tree_3_ref);
}

static void hook(tree)
  tp tree;
{
  if(--tree->walk);
  else if(tree->terminate || tree->del_first || tree->postwalk_root){
    int sv;
    extern Q9000058;		/* ENG_TRAIL_HOOKS ... call Quintus */

    sv = Q9000058;
    Q9000058 = 0;

    if(tree->terminate) avl_delete_tree(tree);
    else{

      QP_term_ref ref = QP_new_term_ref();
      QP_term_ref Order = QP_new_term_ref();
      QP_term_ref key = QP_new_term_ref();
      QP_term_ref term = QP_new_term_ref();

      QP_put_db_reference(ref, tree->tree_3_ref);
      QP_query(order, ref, Order);

      /* take care of nodes to delete after walk completes*/

      if(tree->del_first){
	delete_data dd;
	np next, n;

	dd.ref = ref;
        dd.term = term;
	dd.key = key;
	dd.order = Order;

	next = tree->del_first;
	do{
	  n = next;
	  next = next->del_next;
	  (void) instance(ref, &n->it, key, term);
	  delete_recurse(&tree->root, &dd);
	}while(n != next);	/* last node of list points to self */

	tree->del_first = (np) 0;
      }

      /* take care of nodes to insert after walk completes */

      if(tree->postwalk_root){
	insert_data id;
	id.ref = ref;
	id.key = key;
	id.term = term;
	id.order = Order;
	id.allow_duplicates = tree->allow_duplicates;
	insert_subtree(&tree->root, tree->postwalk_root, &id);
	tree->postwalk_root = (np) 0;
      }
    }
    Q9000058 = sv;
  }
}

typedef void (*fn)();

fn avl_trail(tree)
  tp tree;
{
  tree->walk++;
  return hook;
}

#endif
