#ifndef	lint
static char sccsid[] = "@(#)kdb.c	1.7 29 Sep 1993";
#endif

/* kdb: keyed database operations. This is essentially a collection of
doubly linked queue operations on asserted terms. A common abbreviation
for "doubly linked queues" in algorithms text is "deques."

The Prolog end of this should be in the file "kdb.pl".

I'm using the QP_[free,malloc] operations because, unlike the standard
malloc(), free() operations, they will actually return memory to the
operating system.

Using the QP_ stream I/O functions ensures that the redirections done
under QUI, for example, will work.

This software is free: please copy it and/or modify it. Do with it what
you like.  It is a gift from KnowledgeWare Inc.

Tom Howland, April 1993

*/

#include <quintus/quintus.h>

typedef struct kdb_node {
  struct kdb_node *next, *prev;
  QP_db_reference it;
} *np;

typedef struct {
  np head, tail;
  int count;
} deque;

static QP_pred_ref erase;

/* set up the above static pointer. */

void kdb_init()
{
  erase = QP_predicate("erase", 1, "kdb");
  if(erase == QP_BAD_PREDREF) {
    QP_fprintf(QP_stderr, "kdb_init: could not find kdb:erase/1");
    QP_action(QP_ABORT);
  }
}

/* create a new deque, which is associated with a "key" in Prolog. */

deque * kdb_allocate_deque(){
  deque *D;

  D = (deque *) QP_malloc(sizeof(deque));
  D->head = D->tail = 0;
  D->count = 0;
  return D;
}

/* put one at the front of the deque */

np kdb_a(D, ip)
  deque *D; QP_term_ref ip;
{
  np tp, t;

  tp = (np) QP_malloc(sizeof(struct kdb_node));
  tp->prev = (np) 0;
  if(t = D->head) t->prev = tp, tp->next = t;
  else D->tail = tp, tp->next = (np) 0;
  D->head = tp;
  (void) QP_get_db_reference(ip, &tp->it);
  D->count++;
  return tp;
}

/* put one at the end of the deque */

np kdb_z(D, ip)
  deque *D; QP_term_ref ip;
{
  np tp, t;

  tp = (np) QP_malloc(sizeof(struct kdb_node));
  tp->next = (np) 0;
  if(t = D->tail) t->next = tp, tp->prev = t;
  else D->head = tp, tp->prev = (np) 0;
  D->tail = tp;
  (void) QP_get_db_reference(ip, &tp->it);
  D->count++;
  return tp;
}

/* put one after another */

np kdb_after(D, Fore, Term)
  deque *D; np Fore; QP_term_ref Term;
{
  np Aft;
  np t;

  if(Fore == D->tail) return kdb_z(D, Term);

  Aft = (np) QP_malloc(sizeof(struct kdb_node));
  (void) QP_get_db_reference(Term, &Aft->it);
  t = Fore->next;
  Aft->next = t;
  t->prev = Aft;
  Aft->prev = Fore;
  Fore->next = Aft;
  D->count++;
  return Aft;
}

/* put one before another */

np kdb_before(D, Aft, Term)
  deque *D; np Aft; QP_term_ref Term;
{
  np Fore;
  np t;

  if(Aft == D->head) return kdb_a(D, Term);

  Fore = (np) QP_malloc(sizeof(struct kdb_node));
  (void) QP_get_db_reference(Term, &Fore->it);
  t = Aft->prev;
  Fore->prev = t;
  t->next = Fore;
  Fore->next = Aft;
  Aft->prev = Fore;
  D->count++;
  return Fore;
}

/* replace the database reference with another */

void kdb_replace(I, Term)
  np I; QP_term_ref Term;
{
  (void) QP_get_db_reference(Term, &I->it);
}

/* find the database reference at the given node pointer, if there is
   one. */

int kdb_it(I, ref)
  np I; QP_term_ref ref;
{
  if(I->it.address){
    QP_put_db_reference(ref, I->it);
    return 0;
  } else return -1;
}

/* return the head of the deque, skipping over "erased" terms. */

np kdb_head(D)
  deque *D;
{
  np I;
  for(I = D->head; I && !(I->it.address); I = I->next);
  return I;
}

/* return the tail of the deque, skipping over "erased" terms. */

np kdb_tail(D)
  deque *D;
{
  np I;
  for(I = D->tail; I && !(I->it.address); I = I->prev);
  return I;
}

/* return the next entry in the deque, skipping over "erased" terms. */

np kdb_next(I)
  np I;
{
  for(I = I->next; I && !(I->it.address); I = I->next);
  return I;
}

/* return the previous entry in the deque, skipping over "erased" terms. */

np kdb_prev(I)
  np I;
{
  for(I = I->prev; I && !(I->it.address); I = I->prev);
  return I;
}

/* return the nth entry in the deque, not counting "erased" terms. */

np kdb_nth(D, n)
  deque *D; int n;
{
  int i; np p;

  if(n < 1) return (np) 0;
  for(i = 1, p = kdb_head(D); p && i < n; p = kdb_next(p), i++);
  return p;
}

/* erase a database reference, given a refererence variable */

static void erase_it(it, ref)
  QP_db_reference *it;
  QP_term_ref ref;
{
  QP_put_db_reference(ref, *it);
  (void) QP_query(erase, ref);
}

/* erase the database reference, if there is one, but don't mess with the
deque pointers, just indicate that it has been erased. */

void kdb_erase(D, I)
  deque *D;
  np I;
{
  if(I->it.address){
    erase_it(&I->it, QP_new_term_ref());
    I->it.address = 0;
    D->count--;
  }
}

/* remove the deque pointers for a given entry. */

static void hard_erase(D, I)
  deque *D; np I;
{
  if(I == D->head){

    if(I == D->tail) D->head = D->tail = (np) 0;
    else {

      np n;

      n = I->next;
      D->head = n;
      n->prev = (np) 0;
    }

  } else if(I == D->tail){

    np p;

    p = I->prev;
    D->tail = p;
    p->next = (np) 0;

  } else {

    np n,p;

    n = I->next;
    p = I->prev;

    n->prev = p;
    p->next = n;
  }
  QP_free(I);
}

/* erase the database reference, if there is one, and remove the deque
   pointers. */

void kdb_hard_erase(D, I)
  deque *D; np I;
{
  kdb_erase(D, I);
  hard_erase(D, I);
}

/* remove an entire deque and its related database entries */

void kdb_eraseall(D)
  deque *D;
{
  np p = D->head;
  QP_term_ref ref = QP_new_term_ref();
  while(p){
    np x = p;
    p = p->next;
    if(x->it.address) erase_it(&x->it, ref);
    QP_free(x);
  }
  QP_free(D);
}

/* remove the nodes that contain erased database entries */

void kdb_expunge(D)
  deque *D;
{
  np p;
  p = D->head;
  while(p){
    np x = p;
    p = p->next;
    if(x->it.address);
    else hard_erase(D, x);
  }
}

kdb_key_count(D)
  deque *D;
{
  return D->count;
}
