/* @(#)kdb.pl	1.15 05 Oct 1993

Here we emulate Arity's built-in keyed database operations. This can be
seen as an improvement on the standard Prolog database operations
because here we can go forward or backward in the database without
relying on backtracking or aggregate operators such as findall/3. We
also have complete control over the clause order of the database.

This software is free: please copy it and/or modify it. Do with it what
you like.  It is a gift from KnowledgeWare Inc, and was written by Tom
Howland.

*/

:- module(kdb, [

    kdb_ref/1,                   % +Ref

	% kdb_ref(+Ref) indicates if Ref is the correct form for a keyed
	% database reference. It says nothing about whether it is a
	% valid reference.

/* adding to database keys */

    kdb_recorda/3,		% +Key, +Term, -Ref

	% add a term to the beginning of the chain of terms having the
	% same key. It returns the reference structure that has been
	% assigned to the term.

    kdb_recordz/3,		% +Key, +Term, -Ref

	% adds a term to the end of the chain of terms having the same
	% key. It returns the database reference structure that has been
	% assigned to the term.

    kdb_record_after/3,		% +Ref, +Term, -NewRef
    kdb_record_before/3,	% +Ref, +Term, -NewRef

	% These predicates insert a term into a chain of terms. You
	% specify the term that the new term will follow by supplying
	% its database reference structure with the Ref argument. The
	% Term argument is the new term that you want to add. The
	% predicate returns the reference structure for the new term to
	% the NewRef argument.

    kdb_replace/2,		% +Ref, +Term

	% looks up the term that has the given database reference
	% structure and replaces that term with the one you have
	% specified.

    % not implemented: load_key/[2,3], [begin,end]_choices/1

/* Reviewing Database keys */

    kdb_recorded/3,	% +Key, *Term, *Ref

	% searches the database for the terms stored under a given key.
	% You can backtrack to see all the terms.

    kdb_recorded_tro/3,	% +Key, *Term, *Ref

	% This is the same as kdb_recored/3, except that it is safe to
	% call kdb_hard_erase/1 on the current reference, since it
	% "looks ahead".

    kdb_instance/2,	% +Ref, -Term

	% returns the term that is associated with the specified
	% reference.

    kdb_keys/1,		% *Key

	% returns the most general functor for a particular key.

    kdb_key/2,		% +Key, -Ref

	% returns the structure for the specified key. This may be used
	% as input to kdb_{n,p}ref/2. In fact, it can't be used for
	% anything else.

    kdb_key_count/2,	% +Key, -Count

	% returns a Count of the number of terms stored under a given
	% database Key.

    kdb_nref/2,		% +Ref, -NextRef
    kdb_pref/2,		% +Ref, -PreviousRef

	% returns the reference number of the next or previous term in
	% the chain. You can use these predicates to move forward or
	% backward through a chain of terms. These predicates fail if
	% there is no previous or next term in the database key.

	% Note the following idioms. To get to the head of the list, use
	% "kdb_key(Key, D), kdb_nref(D, Head)". To get to the tail of the
	% list, use "kdb_key(Key, D), kdb_pref(D, Tail)".

    kdb_nth_ref/3,	% +Key, +N, -Ref

	% returns the database reference structure for the term some
	% number of positions away from the top of a chain of terms. The
	% Key argument names the key under which the term is stored. The
	% Nargument gives the number of positions away from the top of
	% the key the desired term can be found. N can be positive or
	% negative.

    kdb_sortkey/1,	% +Key

	% performs a "keysort" on the terms stored under the specified
	% database Key. The terms are sorted in standeard order and
	% replaced in the key.

    % not implemented: recorded_nth/4, recorded_ref/4, recorded_terms/3,
    % key/1, write_key/3

/* removing Terms from the keyed database */

    kdb_erase/1,	% +Ref

	% performs a soft werase removing the term that has the given
	% database reference structrue. The system keeps track of the
	% terms that have been erased with this predicate. These terms
	% can be removed with kdb_expunge/0. In addition, kdb_nref/2,
	% and kdb_pref/2 pointers remain so that the database reference
	% can still be followed.

    kdb_expunge/0,

	% This is necessary when your program uses kdb_erase/1. It frees
	% the references to erased terms. If you have a backtracking
	% procedure which uses kdb_recorded/3 to return a term and
	% kdb_erase/1 and one of the kdb_record{a,z}/3 to erase and add
	% terms, you would not use kdb_expunge/0 until after this
	% procedure was completed. This is because kdb_recoreded/3 uses
	% the references for the term it is returning to determine the
	% reference to the next term.

    kdb_eraseall/1,	% +Key

	% removes all the terms that are stored under the specified key.

    kdb_hard_erase/1,	% +Ref

	% erases the term with the specified reference structure.
	% However, unlike kdb_erase/1, kdb_hard_erase/1 does not keep
	% track of the terms which have been erased. As such, this
	% predicate should only be used if your program is designed so
	% that there is no danger of the system reference a database
	% term that you have previously eliminatedl Such a situation can
	% occur if you have a backtracking procedure which returns a
	% term using kdb_recorded_tro/3 and then erases that term.

    % not implemented: mth_ref/3

    kdb_unique/2]).		 % +Ref, -Integer

	% kdb_unique/2 generates a unique number given a valid kdb
	% database refererence. It is a partial answer to the
	% descrepancy between this implementations, where keyed database
	% references are represented as structures, and Arity's, where
	% they are represented as atoms.

sccsid('@(#)kdb.pl	1.15 05 Oct 1993').

:- initialization kdb_init.

:- extern(erase(+term)).

/**************************************************************

The keyed database key stuff is implemented largely in C in the file
"kdb.c".  It is a collection of operations on doubly linked lists, also
known as "deques" (double ended queues). In LISP this is known as
"property lists."

See library(newqueues) for a much faster way of doing many of the
operations defined here. */

:- dynamic  key/2, it/2.
:- volatile key/2, it/2.

kdb_ref(X) :- nonvar(X), ref(X).

ref(kdb_database_reference(_,_)).
ref(kdb_deque_pointer(_)).

% "return the most general functor for a particular key"

kdb_keys(Key) :- key(Key, _).

/* adding to database keys */

% kdb_recorda(+Key, +Term, -Ref) 

kdb_recorda(Key, Term, kdb_database_reference(D,I)) :-
    record1(Key, Term, D, R),
    kdb_a(D, R, I).

record1(Key, Term, D, R) :-
    clean_key(Key, K),
    (   key(K, D)
    ->  true
    ;   kdb_allocate_deque(D),
        assert(key(K, D))
    ),
    assert_it(Term, R).

/* clean_key/2 has a call to kdb_unique/2, since under Arity database
references are atomic. */

clean_key(Key, K) :-
    (   kdb_unique(Key, K)
    ->  true
    ;   functor(Key, Name, Arity),
        functor(K, Name, Arity)
    ).

assert_it(Term, R) :- assert(it(_,Term), R).	% defeats indexing

% kdb_recordz(+Key, +Term, -Ref) 

kdb_recordz(Key, Term, kdb_database_reference(D,I)) :-
    record1(Key, Term, D, R),
    kdb_z(D, R, I).

% kdb_record_after(+Ref, +Term, -NewRef) 

kdb_record_after(kdb_database_reference(D,Fore), Term,
		 kdb_database_reference(D,Aft)) :-
    assert_it(Term, R),
    kdb_after(D, Fore, R, Aft).

% kdb_record_before(+Ref, +Term, -NewRef) 

kdb_record_before(kdb_database_reference(D,Aft), Term,
		  kdb_database_reference(D,Fore)) :-
    assert_it(Term, R),
    kdb_before(D, Aft, R, Fore).

% kdb_replace(+Ref, +Term) 

kdb_replace(kdb_database_reference(_,I), Term) :-
    fetch(I, It),
    erase(It),
    assert_it(Term, R),
    c_kdb_replace(I, R).

fetch(I, It) :- kdb_it(I, It, 0).

% kdb_sortkey(+Key)

kdb_sortkey(Key) :-
    deque(Key, D),
    kdb_head(D, I),
    fingy(I, L),
    sort(L, Sorted),
    replace_them(L, Sorted).

fingy(I, L) :-
    (   I =:= 0
    ->  L = []
    ;   L = [x(Term,I,R)|T],
        kdb_it(I, R, _), clause(it(_, Term), _, R), kdb_next(I, J),
	fingy(J, T)
    ).

replace_them([], _).
replace_them([x(_,I,_)|T], [x(_,_,R)|Ts]) :-
    c_kdb_replace(I, R),
    replace_them(T, Ts).

/***************************************************************

Reviewing Database keys */

% kdb_recorded(+Key, *Term, *Ref) 

kdb_recorded(Key, Term, kdb_database_reference(D,I)) :-
    (   nonvar(I)
    ->  instance_it(I, Term)
    ;   nonvar(Key), deque(Key, D), head(D, Ii), linear(Ii, Term, I)
    ).

deque(Key, D) :-
    clean_key(Key, K),
    key(K, D).

linear(I, Term, I) :- instance_it(I, Term).
linear(Ii, Term, I) :- next(Ii, N), linear(N, Term, I).

instance_it(I, Term) :- fetch(I, R), clause(it(_, Term), _, R).

next(I, N) :- kdb_next(I, N), N =\= 0.

% kdb_recorded_tro(+Key, +Term, -Ref) 

kdb_recorded_tro(Key, Term, kdb_database_reference(D,I)) :-
    (   nonvar(I)
    ->  instance_it(I, Term)
    ;   nonvar(Key), deque(Key, D), head(D, Ii), kdb_next(Ii, N),
        linear(Ii, N, Term, I)
    ).

linear(I, _, Term, I) :- instance_it(I, Term).
linear(_, N, Term, I) :- N =\= 0, kdb_next(N, M), linear(N, M, Term, I).

% kdb_instance(+Ref, -Term) 

kdb_instance(kdb_database_reference(_,I), Term) :- instance_it(I, Term).

% kdb_key(+Key, -Ref) 

kdb_key(Key, kdb_deque_pointer(D)) :- deque(Key, D).

% kdb_key_count(+Key, -Count)

kdb_key_count(Key, Count) :-
    deque(Key, D),
    c_kdb_key_count(D, Count).

% kdb_nref(+Ref, -NextRef) 

kdb_nref(kdb_deque_pointer(D), kdb_database_reference(D, I)) :- head(D, I).
kdb_nref(kdb_database_reference(D,Ii), kdb_database_reference(D,I)) :-
    next(Ii, I).

% kdb_pref(+Ref, -PreviousRef) 

kdb_pref(kdb_deque_pointer(D), kdb_database_reference(D, I)) :- tail(D, I).
kdb_pref(kdb_database_reference(D,Ii), kdb_database_reference(D,I)) :-
    kdb_prev(Ii, I), I =\= 0.

head(D, I) :- kdb_head(D, I), I =\= 0.

tail(D, I) :- kdb_tail(D, I), I =\= 0.

% kdb_nth_ref(+Key, +N, -Ref) 

kdb_nth_ref(Key, N, kdb_database_reference(D,I)) :-
    deque(Key, D), kdb_nth(D, N, I), I =\= 0.

/***************************************************************

removing Terms from the keyed database */

% kdb_erase(+Ref) 

kdb_erase(kdb_database_reference(D,I)) :- kdb_erase(D, I).

% kdb_hard_erase(+Ref)

kdb_hard_erase(kdb_database_reference(D,I)) :- c_kdb_hard_erase(D, I).

% kdb_eraseall(+Key)

kdb_eraseall(Key) :-
    (   var(Key)
    ->  true
    ;   clean_key(Key, K), retract(key(K, D))
    ->  c_kdb_eraseall(D)
    ;   true
    ).

% kdb_expunge

kdb_expunge :-
    key(_, D),
    c_kdb_expunge(D),
    fail ; true.

% kdb_unique(+Ref, -Int)

kdb_unique(Ref, Int) :- ground(Ref), unique(Ref, Int).

unique(kdb_database_reference(_,I), I).
unique(kdb_deque_pointer(D), D).

%%%%%%%%%%%

foreign(kdb_a, c, kdb_a(+address, +term, [-address])).
foreign(kdb_after, c, kdb_after(+address, +address, +term, [-address])).
foreign(kdb_allocate_deque, c, kdb_allocate_deque([-address])).
foreign(kdb_before, c, kdb_before(+address, +address, +term, [-address])).
foreign(kdb_erase, c, kdb_erase(+address, +address)).
foreign(kdb_eraseall, c, c_kdb_eraseall(+address)).
foreign(kdb_expunge, c, c_kdb_expunge(+address)).
foreign(kdb_hard_erase, c, c_kdb_hard_erase(+address, +address)).
foreign(kdb_head, c, kdb_head(+address, [-address])).
foreign(kdb_init, c, kdb_init).
foreign(kdb_it, c, kdb_it(+address, -term, [-address])).
foreign(kdb_key_count, c, c_kdb_key_count(+address, [-integer])).
foreign(kdb_next, c, kdb_next(+address, [-address])).
foreign(kdb_nth, c, kdb_nth(+address, +integer, [-address])).
foreign(kdb_prev, c, kdb_prev(+address, [-address])).
foreign(kdb_replace, c, c_kdb_replace(+address, +term)).
foreign(kdb_tail, c, kdb_tail(+address, [-address])).
foreign(kdb_z, c, kdb_z(+address, +term, [-address])).

foreign_file(system(libbuiltins_d), [
	     kdb_a,
	     kdb_after,
	     kdb_allocate_deque,
	     kdb_before,
	     kdb_erase,
	     kdb_eraseall,
	     kdb_expunge,
	     kdb_hard_erase,
	     kdb_head,
	     kdb_init,
	     kdb_it,
	     kdb_key_count,
	     kdb_next,
	     kdb_nth,
	     kdb_prev,
	     kdb_replace,
	     kdb_tail,
	     kdb_z]).

:- load_foreign_executable(system(libbuiltins_d)), abolish(foreign_file, 2),
    abolish(foreign, 3).

end_of_file.

/* test code */

:- use_module(library(basics)).
:- use_module(builtins(kdb)).

init_db :-
    member(X, [a,b,c]),
    kdb_recordz(dir, X, _),
    fail ; true.

forward :-
    kdb_key(dir,KeyRef),
    kdb_nref(KeyRef, Ref),
    go_forward(Ref).

go_forward(Ref) :-
    kdb_instance(Ref,Term),
    write(Term),nl,
    (   kdb_nref(Ref,Nref)
    ->  go_forward(Nref)
    ;   true
    ).

backward :-
    kdb_key(dir, KeyRef),
    kdb_pref(KeyRef, Ref),
    go_backward(Ref).

go_backward(Ref):-
    kdb_instance(Ref,Term),
    write(Term),nl,
    (   kdb_pref(Ref,Pref)
    ->  go_backward(Pref)
    ;   true
    ).

eek :-
    kdb_eraseall(dir),
    init_db.

kee :- forward, backward.

rma :- eek, kdb_recorded(dir, a, R), kdb_erase(R), kee.
rmb :- eek, kdb_recorded(dir, b, R), kdb_erase(R), kee.
rmc :- eek, kdb_recorded(dir, c, R), kdb_erase(R), kee.

sa :- eek, kdb_recorded_tro(dir, a, R), kdb_hard_erase(R), kee.
sb :- eek, kdb_recorded_tro(dir, b, R), kdb_hard_erase(R), kee.
sc :- eek, kdb_recorded_tro(dir, c, R), kdb_hard_erase(R), kee.

% the following, b/0, works, although Arity says it breaks. It works
% because even though it is referencing deallocated nodes, the nodes are
% still there, their memory hasn't been used by something else. So it
% shouldn't be counted on working, just like Arity's implementation.

b :- eek, kdb_recorded(dir, T, R),
    format('extracted ~q~n', [T]),
    kdb_hard_erase(R),
    fail ; true.

% the Arity advocated method:

a :- eek, kdb_recorded(dir, T, R),
    format('extracted ~q~n', [T]),
    kdb_erase(R),
    fail ; true.

% c/0 should break, like b/0, but doesn't, for the same reasons

c :- eek, kdb_recorded(dir, T, R),
    format('extracted ~q~n', [T]),
    kdb_erase(R), kdb_expunge,
    fail ; kdb_expunge.

ra :- eek, kdb_recorded(dir, a, R), kdb_replace(R, foo), kee.
rb :- eek, kdb_recorded(dir, b, R), kdb_replace(R, foo), kee.
rc :- eek, kdb_recorded(dir, c, R), kdb_replace(R, foo), kee.

% load the database backwards

init_dba :-
    kdb_eraseall(dir),
    member(X, [a,b,c]),
    kdb_recorda(dir, X, _),
    fail ; kee.

% test record_[after,before]

ta :- eek, kdb_recorded(dir, a, R), kdb_record_after(R, foo, _),
    kee.
tal :- eek, kdb_key(dir, Key), kdb_pref(Key, Last),
    kdb_record_after(Last, foo, _), kee.
tb :- eek, kdb_recorded(dir, c, R), kdb_record_before(R, foo, _),
    kee.
tbf :- eek, kdb_key(dir, Key), kdb_nref(Key, First),
    kdb_record_before(First, foo, _), kee.

% test nth.

n(X) :-
    kdb_nth_ref(dir, X, Ref),
    kdb_instance(Ref, Y),
    format('reference #~d is ~a', [X, Y]).

n :- eek, kdb_recorded(dir, b, R), kdb_erase(R), n(2).
n2 :- eek, kdb_recorded_tro(dir, b, R), kdb_hard_erase(R), n(2).
