/* @(#)tavl.pl	1.18 05 Oct 1993

tavl ... Tom's avl package (to distinguish it from library(avl))

Here we emulate Arity's b-tree database operations.

This software is free: please copy it and/or modify it. Do with it what
you like.  It is a gift from KnowledgeWare Inc, and was written by Tom
Howland.

*/

:- module(tavl, [

/* Adding terms to a b-tree */

    defineb/3,		% +BTree, +Uniqueness, +Order

	/* This predicate can be ignored. It allows you to specify a
	   funny sorting order and whether or not duplicates are
	   allowed.

	   Creates a b-tree named BTree, with the given order, Order.
	   The argument BTree must be a ground term. The Order argument
	   can be used to specify the sorting order. The default for
	   Order is '+', which means sort the entire term in standard
	   order. An Order argument such as +(-,+) means "sort the
	   principal functor in standard order, the first argument in
	   reverse standard order, and the last argument in standard
	   order". So you can have order arguments like -(+,-,-(+,-)),
	   etc.

	   If Uniqueness = true, then duplicates are not allowed in the
	   b-tree. Duplicates are allowed by default.
	*/

    recordb/3,		% +Btree, +Key, +Term

	/* Saves the Key and its associated Term under the given Btree */

    replaceb/4,		% +BTree, +Key, +Old, +New

	/* Replaces the term Old saved under Key with the new term New */

/* Reviewing terms in a B-tree */

    retrieveb/3,		% +BTree, ?Key, ?Term

	/* Fetches the Keys and Terms from the Btree according to Order */

    retrieveb_nth/3,		% +BTree, +N, -Key

	/* Fetches the nth Key from the Btree */

    btree_count/2,		% +BTree, -Count

	/* Returns how many elements are in the given Btree */

    % not implemented: retrieveb_terms/3, betweenb/7, betweenkeysb/4,
    % retrieveb_keys/2, retrieve_terms/2
    % what_btrees/1

/* Removing terms from a B-tree */

    removeb/2,		% +BTree, +Key

	/* Removes all elements stored under Key in BTree */

    removeb/3,		% +BTree, +Key, +Term

	/* Backtracks. Removes the given Key and Term from Btree. */

    removeallb/1]).	% +BTree

	/* Frees a Btree */

/* NOTE: this implementation is just as broken as Arity's. That is,
modifications to a btree during a "walk" (while a call to retrieveb is
active) will have unpredictable results. The following discussion talks
about how the use of a special hack from Quintus made this work much
better than Arity's, but that hack doesn't work under Solaris 2.2, so it
has been commented out. Quintus has hinted that it might support this
hack in release 3.2 (Fall '93), so we'll wait and see. */

/* The btree predicates.

Here is a summary of what we have.

         | avl trees
--------------------
insert   | O(log N)
delete   | O(log N)
find     | O(log N)
walk     | O(N)

After struggling with this C implementation of avl trees for a week, I
finally realized that a satisfactory implementation cannot be realized.
Perhaps my criteria for satisfaction is too high.

What's needed is a way of arranging that a clean up predicate will be
called whenever a walk of the tree is terminated. In Prolog this means
whenever retrieveb/3 is determinately exited. There are three
kinds of determinate exits: an abort, a cut by some parent goal, or a
failure. The Prolog system only lets the programmer catch an exit by
failure.

Quintus supplies an undocumented predicate for this very situation. 
'Quintus: trail'(Hook, Arg) arranges for the C function Hook to be
called with the argument Arg upon determinate exit of the containing
clause.  But at this point the trail hook can not call Prolog.  But in
order to do insertions or deletions on a btree, we must call Prolog to
do the comparisons.  So we're screwed.  This is a bug in Prolog, I
refuse to be blamed.  Sictus Prolog supplies an "undo" predicate.  This
is precisely what is needed here.  I submitted a bug report to Quintus.

Quintus supplied me with another undocumented and unsupported feature
that will allow me to call Prolog from a trail hook. This involves
setting the symbol Q9000058 to zero upon entry to the trail hook, and
then setting it back to its original value before leaving the trail
hook.  This symbol is known to Quintus (Tim Lindholm, in particular) as
"ENG_TRAIL_HOOKS". Tim says that maybe the trail hooks will be a
supported feature by release 3.2. In any case, if it isn't, "avl.c" will
have to be updated.

What this means for you the programmer is that whenever you call
retrieveb/3, you should either call it exhaustively (in a fail
driven loop) or use the idiom

  retreiveb(BTree, Key, Term),
  !,
  somthing(BTree, Key, Term).

In other words, if "something/3" calls anything that may modify the
btree, those modifications will not be visible until retrieveb/3
has been determinately exited and then some new call has been made to
the btree predicates.  This is a feature.  Without it, we wouldn't be
able to walk the tree.

Why wouldn't we be able to walk the tree? Because the walk expects node
pointers to be unchanging. If we allowed modifications (insertions or
deletions) during the walk, where would our pointers take us?

Which is not to say that we can't do deletions and insertions during a walk.
We can. I went to a great deal of trouble to ensure that we can. They simply
won't be visible until the walk completes. This has an analog: the "logical
view" of database updates. The "logical view" is described in the
Quintus manual in the introduction to the chapter on the database.

This implementation struggles to be semantically and operationally
identical to the Arity supplied primitives. It deviates in the following
ways:

  (1) there is no checking for ground terms and keys. This interacts
      with the +Uniqueness directive to defineb/3: it produces a
      subsumption check.

      Asserting a general key or term will block out assertions of more
      specific keys or terms. Likewise an asertion of a general key or
      term will be blocked out by the assertion of a more specific key
      or term. But only when the Uniqueness switch is on. It's off by
      default.

  (2) Arity allows you to define an "order" parameter. And then lets you
      deviate from it. I don't let you deviate from it. If you supplied
      an Order parameter, I expect you to use it. This isn't because I'm
      a style nazi, but because I don't feel like programming more of a
      kludge if I can avoid it. Another reason is that I can supply a
      more compact representation if I allow no deviations. Someone has
      to say to me "But Tom! We MUST be able to deviate from the order
      parameter!" -- not likely.

      What does it mean to "deviate from the order parameter?" Suppose
      you supply an order parameter +(-). This means you will be using
      an arity 1 term, sorted in ascending order on the principal
      functor, and in descending order on the first argument. I went to
      the trouble of implementing this when to my horror I found that
      Arity will then let you use the same btree to store terms of any
      form, and that Arity will continue to apply the order directive to
      the terms, using standard order when the term begins to diverge
      from the pattern. This is too much like A.I. for me! Wow!

  (3) If you delete something during a walk with Arity's implementation
      and Arity walks over it, it aborts with the message "illegal database
      access" or some such rubbish. This implementation saves the
      modifications until after the walk completes.

  (4) The BTree name parameter must be an atom in Arity's implementation. In
      this implementation, it must be a ground term.


Another style note.  Since removeb/3 instantiates its Key and Term
arguments but doesn't leave a choice point, you might think that an
ideal way of traversing the tree is

	repeat,
	    (   removeb(BTree, Key, Term)
	    ->  process(Key, Term)
	    ;   !, fail
	    )

Ideal, that is, if you're deleting things in order. I know it's ugly,
please don't tell me. The problem with this is that it requires a lookup
of the smallest element each time removeb/3 is called. The advantage is
that the thing is deleted immediately. Conversely, the form

	retrieveb(BTree, Key, Term),
	process(Key, Term),
	removeb(BTree, Key, Term),
	fail ; true

is just as slow as the previous form, but the modifications to the btree
will not be apparent to process/2 until the final element of the btree
has been found. Confusing, isn't it.

You might be saying to yourself: but why use avl's, why not 2-3 trees or
something else? As far as I know, there is no algorithm that limits
comparisons as well as avl.  If there is, tell me.  Comparisons are the
most expensive part here.  2-3 trees are optimal in limiting disk
access.  But all our stuff is in memory. 

So I think that avl is a pretty good idea?  No. The data should be
organized to exploit first argument indexing.  I can't believe that that
isn't possible.  An even better solution would be to not use the
asserted database at all, instead use library(avl). I did some timings
which indicate that library(avl) is roughly an order of magnitude faster
than this. Unfortunately, library(avl) isn't complete: the hardest part,
deletion, was never written.

*/

sccsid('@(#)tavl.pl	1.18 05 Oct 1993').

:- initialization avl_init.

:- dynamic  tree/3, bt/3.
:- volatile tree/3, bt/3.

/* Adding terms to a b-tree */

% defineb(+BTree, +Uniqueness, +Order)

defineb(Tag, Unique, Order) :-
    ground(Tag),
    (   tree(Tag, _, _)
    ->  (   retrieveb(Tag, _, _)
	->  fail
	;   retractall(tree(Tag, _, _)), defineb(Tag, Unique, Order)
	)
    ;   (   dup(Unique, Dup) -> true  ),
        avl_new_tree(Dup, Tree),
	assert(tree(Tag, Tree, Order))
%	assert(tree(Tag, Tree, Order), Ref),
%       avl_save_tree_ref(Tree, Ref)
    ).

dup(false, -1).
dup(true, 0).

% recordb(+Btree, +Key, +Term)

recordb(Tag, Key, Term) :-
    (   is_tree(Tag, Tree, Order)
    ->  avl_insert(Tree, Order, Key, Term, 0)
    ;   defineb(Tag, false, +), recordb(Tag, Key, Term)
    ).

is_tree(Tag, Tree, Order) :- ground(Tag), tree(Tag, Tree, Order).

/* bt_instance/3 is called by avl_delete() */

:- extern(bt_instance(+term, -term, -term)).
bt_instance(Ref, Key, Term) :- clause(bt(_,Key, Term), _, Ref).

:- extern(bt_unify(+term, +term, +term)).
bt_unify(Ref, Key, Term) :- clause(bt(_,Key, Term), _, Ref).

/* bt_assert/3 and ref_key/2 are called by avl_insert() */

:- extern(bt_assert(+term, +term, -term)).
bt_assert(Key, Term, Ref) :- assert(bt(_, Key, Term), Ref).

:- extern(ref_key(+term, +term)).
ref_key(Ref, Key) :- bt_instance(Ref, Key, _).

% replaceb(+BTree, +Key, +Old, +New)

replaceb(Tag, Key, Old, New) :-
    removeb(Tag, Key, Old),
    recordb(Tag, Key, New).

:- extern(erase(+term)).

/* Reviewing terms in a B-tree */

    % not implemented: retrieveb_terms/3, betweenb/7, betweenkeysb/4,
    % retrieveb_keys/2, retrieve_terms/2, btree_count/2,
    % what_btrees/1

% retrieveb(+BTree, ?Key, ?Term)

retrieveb(Tag, Key, Term) :-
    is_tree(Tag, Tree, Order),
%    avl_trail(Tree, Hook),
%    '$Quintus: trail'(Hook, Tree),
    avl_find(Tree, Order, Key, Term, Ref, Left, Right, 0),
    next(Left, Right, Ref, More, Key, Term),
    (   More =:= 0, !		% note that this check must occur here
				% in order to force the trail hook,
				% otherwise '$Quintus: trail'/2 leaves a
				% choice point and the hook won't be run
				% until possibly much later.
    ;   true
    ).

next(Left, Right, Ref, More, Key, Term) :-
    (   Left =:= Right
    ->  bt_instance(Ref, Key, Term), More = 0
    ;   bt_instance(Ref, Key, Term), More = 1
    ;   avl_next(Left, R, I), next(I, Right, R, More, Key, Term)
    ).

/*
% order/2 is called by the trail hook.

:- extern(order(+term, -term)).
order(Ref, Order) :- clause(tree(_,_,Order), _, Ref).
*/

% compare_ref_ref/4 is called by avl_find().

:- extern(compare_ref_ref(+term, +term, +term, -integer)).
compare_ref_ref(Order, Ref, R2, Result) :-
    bt_instance(R2, K, T),
    compare_ref_term(Order, Ref, K, T, Result).

% compare_ref_term_vg/5 is called by avl_find (vg ... vars greater)

:-extern(compare_ref_term_vg(+term, +term, +term, +term, -integer)).

compare_ref_term_vg(Order, Ref, Rk, Rt, Result) :-
    compare1(Order, Ref, Rk, Rt, R),
    to_int_vg(R, Result).

to_int_vg(<, -1). to_int_vg(=, 0). to_int_vg(>, 1).
to_int_vg(varl, 1). to_int_vg(varr, -1).

compare1(Order, Ref, Rk, Rt, R) :-
    bt_instance(Ref, Lk, Lt),
    compare(Order, Lk, Rk, R1),
    (   R1 == =
    ->  var_eq(Lt, Rt, R)
    ;   R = R1
    ).

/* compare_ref_term/5 is called by the avl routines. */

:- extern(compare_ref_term(+term, +term, +term, +term, -integer)).

compare_ref_term(Order, Ref, Rk, Rt, Result) :-
    compare1(Order, Ref, Rk, Rt, R),
    to_int(R, Result).

/* compare_key/4 is called by avl_insert when duplicates are not allowed. */

:- extern(compare_key(+term, +term, +term, -integer)).

compare_key(Order, Ref, Rk, Result) :-
    ref_key(Ref, Lk),
    compare(Order, Lk, Rk, R),
    to_int(R, Result).

to_int(<, -1). to_int(=, 0). to_int(>, 1). to_int(varl, -1). to_int(varr, 1).

compare(+, L, R, C) :- !, var_eq(L, R, C).
compare(-, L, R, C) :- !, invert(L, R, C).
compare(P, L, R, C) :-
    (   vars(L, R, C)
    ->  true
    ;   functor(P, Np, A),
        functor(L, Nl, A),
        functor(R, Nr, A),
        compare(Np, Nl, Nr, Nc),
        (   Nc == =
        ->  ca(0, A, P, L, R, C)
        ;   C = Nc
        )
    ).

/* Make sure that variables are considered equal */

vars(L, R, C) :-
    (   var(L)
    ->  (   var(R)
        ->  C = =
	;   C = varl
	)
    ;   var(R)
    ->  C = varr
    ).

var_eq(L, R, C) :-
    (   vars(L, R, C)
    ->  true
    ;   functor(L, Nl, Al), functor(R, Nr, Ar),
        (   neq(Al, Ar, C)
	->  true
	;   neq(Nl, Nr, C)
	->  true
	;   var_eq(0, Al, L, R, C)
	)
    ).

neq(L, R, C) :- compare(C, L, R), C \== = .

var_eq(A, A, _, _, =) :- !.
var_eq(I, A, L, R, C) :-
    J is I+1,
    arg(J, L, Al),
    arg(J, R, Ar),
    var_eq(Al, Ar, C1),
    (   C1 == =
    ->  var_eq(J, A, L, R, C)
    ;   C = C1
    ).

/* make sure that variables are always considered "less than" ground things
   in the tree, even though the search order is inverted. Otherwise we
   wouldn't be able to walk the tree. */

invert(L, R, C) :-
    (   vars(L, R, C)
    ->  true
    ;   functor(L, Nl, Al), functor(R, Nr, Ar),
        (   neq(Ar, Al, C)	% notice that the arguments are reversed
	->  true
	;   neq(Nr, Nl, C)	% ditto
	->  true
	;   invert(0, Al, L, R, C)
	)
    ).

invert(A, A, _, _, =) :- !.
invert(I, A, L, R, C) :-
    J is I+1,
    arg(J, L, Al),
    arg(J, R, Ar),
    invert(Al, Ar, C1),
    (   C1 == =
    ->  invert(J, A, L, R, C)
    ;   C = C1
    ).

ca(A, A, _, _, _, =) :- !.
ca(I, A, P, L, R, C) :-
    J is I+1,
    arg(J, P, Ap),
    arg(J, L, Al),
    arg(J, R, Ar),
    compare(Ap, Al, Ar, C1),
    (   C1 == =
    ->  ca(J, A, P, L, R, C)
    ;   C = C1
    ).

% retrieveb_nth(+BTree, +N, -Key)

retrieveb_nth(Tag, N, Key) :-
    is_tree(Tag, Tree, _),
    avl_nth(Tree, N, Ref, 0),
    ref_key(Ref, Key).

% btree_count(+BTree, -Count)

btree_count(Tag, C) :- is_tree(Tag, Tree, _), avl_count(Tree, C).

/* Removing terms from a B-tree */

% removeb(+BTree, +Key)

% notice that Key need not be bound.

removeb(Tag, Key) :-
    is_tree(Tag, Tree, Order),
    repeat,
        (    delete(Tree, Order, Key, _)
	->   fail
	;    !
	).

delete(Tree, Order, Key, Term) :-
    avl_delete(Tree, Order, Key, Term, 0).

% removeb(+BTree, +Key, +Term)

% neither Key nor Term need be bound.

removeb(Tag, Key, Term) :-
    is_tree(Tag, Tree, Order),
    delete(Tree, Order, Key, Term).

% removeallb(+BTree)

removeallb(Tag) :-
    (   is_tree(Tag, Tree, _)
    ->  avl_delete_tree(Tree),
        retractall(tree(Tag, Tree, _))
    ;   true
    ).

foreign(avl_count, c, avl_count(+address(tt), [-integer])).
foreign(avl_delete, c, avl_delete(+address(tt), +term, +term, +term,
				  [-integer])).
foreign(avl_delete_tree, c, avl_delete_tree(+address(tt))).
foreign(avl_find, c, avl_find(+address(tt), +term, +term, +term, -term,
			      -address(nt), -address(nt), [-integer])).
foreign(avl_init, c, avl_init).
foreign(avl_insert, c, avl_insert(+address(tt), +term, +term, +term,
				  [-integer])).
foreign(avl_new_tree, c, avl_new_tree(+integer, [-address(tt)])).
foreign(avl_next, c, avl_next(+address(nt), -term, [-address(nt)])).
foreign(avl_nth, c, avl_nth(+address(tt), +integer, -term, [-integer])).
% foreign(avl_save_tree_ref, c, avl_save_tree_ref(+address(tt), +term)).
% foreign(avl_trail, c, avl_trail(+address(tt), [-address])).

foreign_file(system(libbuiltins_d), [
             avl_count,	
	     avl_delete,
	     avl_delete_tree,
	     avl_find,
	     avl_init,
	     avl_insert,
	     avl_new_tree,
	     avl_next,
%	     avl_trail,
%	     avl_save_tree_ref,
	     avl_nth]).

:- load_foreign_executable(system(libbuiltins_d)), abolish(foreign_file, 2),
    abolish(foreign, 3).


end_of_file.

/* Test code */

:-use_module(library(avl)).
:-use_module(library(basics)).
:-use_module(library(between)).
:-use_module(library(random)).
:-use_module(library(lists)).
:-use_module(builtins(tavl)).

idb :- 
    removeallb(student),
    member(X, [student(ralph,d),
	       student(susan,a),
	       student(adam,f),
	       student(linda,b),
	       student(dean,c)]),
    arg(1, X, Key),
    recordb(student, Key, X),
    fail ; true.

bv :- retrieveb(student, Key, Term),
    format('key(~q), term(~q)~n', [Key, Term]),
    fail ; true.

% requires trail hook:

br :- idb,
    retrieveb(student, Key, Term),
    removeb(student, Key, Term),
    format('key(~q), term(~q)~n', [Key, Term]),
    fail ; true.

% test nth
 
nth(N,X) :- retrieveb_nth(student, N, X).
 
nth :- idb, nth(1,adam),nth(2,dean),nth(3,linda),nth(4,ralph),nth(5,susan).

% test btree_count

btc :- idb, btree_count(student, 5).

% unify

t3(X,Y) :- idb, Z=student(linda,X), retrieveb(student, Y, Z).

% make sure that removeb is unifying its deletions

t4(X) :- idb, Z=student(linda,X), removeb(student, _, Z).

% test the incredible order argument

tb3 :-
    removeallb(grades),
    defineb(grades, true, -(+,+)),
    member(X, ['1'(jones,a),'1'(james,d),'1'(barnes,b),'2'(tyler,f),
	       '2'(tyler,c),'3'(regis,a),'1'(barnes,b),'4'(daniels,b)]),
    recordb(grades, X, X),
    fail ;
    retrieveb(grades, X, X),
    format('~N~q', [X]),
    fail ; true.

% subsumption.  Should be only one "jones"

tb4 :-
    removeallb(grades),
    defineb(grades, true, -(+,+)),
    member(X, ['1'(jones,_),'1'(jones,a),'1'(james,d),'1'(barnes,b),'2'(tyler,f),
	       '2'(tyler,c),'3'(regis,a),'1'(barnes,b),'4'(daniels,b)]),
    recordb(grades, X, X),
    fail ;
    retrieveb(grades, X, X),
    format('~N~q', [X]),
    fail ; true.

% intermediate non-matches

tb5 :- 
    removeallb(x),
    member(X, [b(c,d),b(d,x),b(e,d)]),
    recordb(x, X, X),
    fail ;
    X = b(_,d), retrieveb(x, X, X),
    format('~N~q', [X]),
    fail ; true.

% insertions during walk

rm(B, K) :- removeb(B, K).
rm(B) :- removeallb(B).
rm1(B, K) :- removeb(B, K, K).
rt(B, X) :- retrieveb(B, X, X).
rc(B, X) :- recordb(B, X, X).

tb6 :-
    rm(x),
    rc(x, x), rc(x, y),
    findall(Y-Z, foo(Y,Z), L),
    format('~q~n',[L]),
    findall(Z, retrieveb(x, Z, Z), M),
    format('~q~n',[M]).

foo(Y,Z) :-
    rt(x, Z),
    member(Y, [a,b]),
    rc(x, Y).

l(L, L, []) :- !.
l(I, L, [I|T]) :- J is I+1, l(J, L, T).

/* fingy(10000) is a good test for a large tree. First it inserts 10000
   random elements into a tree, then it removes 5000 of them. */

fingy :- fingy(10000).

fingy(Lim) :-
    rm(x),
    l(0, Lim, L),
    random_permutation(L, R),
    (   member(X, R), rc(x, X),	fail
    ;   true
    ),
    (   findall(X, rt(x, X), L)
    ->  true
    ;   raise_exception('INSERTION TEST FAILED!!!')
    ),
    Half is Lim >> 1,
    length(HL, Half),
    random_permutation(R, R2),
    append(HL, HR, R2),
    (   member(X, HL), rm(x, X), fail
    ;   true
    ),
    sort(HR, Sorted),
    (   findall(X, rt(x, X), Sorted)
    ->  true
    ;   raise_exception('DELETION TEST FAILED!!!')
    ),
    rm(x).

fingout :-
    current_output(CO),
    between(1, 9, X),
        fingy(999),
	write(X), write(','), flush_output(CO),
    fail ; true.

/* a/1 is interesting to compare with fingy/1 ... a/1 is an order of
magnitude faster at insertion than "tavl". Unfortunately, Richard never
finished library(avl) : there are no "delete" predicates. */

a(Lim) :-
    pairs(0, Lim, L),
    random_permutation(L, R),
    list_to_avl(R, A),
    (   avl_to_list(A, L)
    ->  true
    ;   format('~NTEST FAILED!!!', [])
    ).

pairs(L, L, []) :- !.
pairs(I, L, [I-I|T]) :- J is I+1, pairs(J, L, T).

/* perm/1 is useful for finding the smallest tree that will make it crash.
   It writes its results in a little local file "perm_log". Since
   perm(9) would run forever (effectively), it overwrites its file.
   Therefore, this test should be run on a local file system to avoid
   abusing nfs. */

perm(Lim) :-
    perm_head(perm_log, CO, S),
    (   between(1, Lim, Count),		% backtracks
	perm_(Count, S),
	fail
    ;   true
    ),
    perm_tail(CO, S).

perm_head(F, CO, S) :-
    current_output(CO),
    open(F, write, [binary, seek(byte)], S),
    set_output(S).
    
perm_(Count, S) :-
    (	l(0, Count, L),
	permutation(L, R),		% backtracks
	seek(S, 0, bof, _),
	format('R=~q, HL=[',[R]),
	stream_position(S, Pos),
	subseq(R, Sub, HR),		% backtracks
	sort(HR, Sorted),
	permutation(Sub, HL),		% backtracks
	rm(x),
	(   member(X, R), rc(x, X), fail
	;   true
	),
	stream_position(S, _, Pos),
	(   member(X, HL), write(X), put(","), flush_output(S), rm(x, X), fail
	;   true
	),
	put("]"), nl, flush_output(S),
	(   findall(X, rt(x, X), Sorted)
	->  true
	;   raise_exception('TEST FAILED!!!'(R, HL))
	),
	fail
    ;   true
    ).

perm_tail(CO, S) :-
    set_output(CO),
    close(S).

perm1(Len) :-
    perm_head(perm_log1, CO, S),
    perm_(Len, S),
    perm_tail(CO, S).

% These two test cases ate up large chunks of my time. It was due to an
% over-eager deletion rebalancing. A one-liner change.

a:-
        R=[0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15], 
        HL=[11,13,15,10,12,8],
	current_output(CO),
	rm(x),
	(   member(X, R), rc(x, X), fail
	;   true
	),
	(   member(X, HL), write(X), put(32), flush_output(CO),
	    removeb(x, X, X), fail
	;   true
	).

b:-
        R=[0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16], 
        HL=[11,12,13,10,8],
	current_output(CO),
	rm(x),
	(   member(X, R), rc(x, X), fail
	;   true
	),
	(   member(X, HL), write(X), put(32), flush_output(CO),
	    removeb(x, X, X), fail
	;   true
	).

/* this one was due to the fact that find_range() was not setting the
   reference to the leftmost occurance. */

test42(X, Y) :-
    recordb(m, t(788,0,1), t(969,0,1)),
    recordb(m, t(793,0,1), t(965,0,1)),
    recordb(m, t(863,0,1), t(961,0,1)),
    retrieveb(m, X, Y).

t42(X) :- retrieveb(m, t(793,0,1), X).
